@@ -1,8 +1,9 @@
 import numpy as np
 from .exceptions import *
-from pathlib2 import Path
+from pathlib import Path
 import re
 import ast
+from collections import OrderedDict
 
 SUPPORTED_VERSIONS = ['4.24', '5.0', '5.00 Bruker JCAMP library', '5.00 BRUKER JCAMP library', '5.01']
 GRAMMAR = {
@@ -15,6 +16,7 @@ GRAMMAR = {
         'SIZE_BRACKET': '^\([^\(\)<>]*\)(?!$)',
         'LIST_DELIMETER': ', ',
         'EQUAL_SIGN': '=',
+        'SINGLE_NUMBER':'-?[\d.]+(?:e[+-]?\d+)?',
         'PARALLEL_BRACKET': '\) ',
         'GEO_OBJ': '\(\(\([\s\S]*\)[\s\S]*\)[\s\S]*\)',
         'HEADER':'TITLE|JCAMPDX|JCAMP-DX|DATA TYPE|DATATYPE|ORIGIN|OWNER',
@@ -93,6 +95,60 @@ class Parameter(object):
         else:
             return False
 
+    @property
+    def tuple(self):
+        value = self.value
+        if isinstance(value, int) or isinstance(value, float):
+            return (value,)
+        else:
+            return tuple(value)
+
+    @property
+    def list(self):
+        value = self.value
+        if isinstance(value, list):
+            return value
+        elif isinstance(value, float) or isinstance(value, int) or isinstance(value, str):
+            return [value]
+        else:
+            return list(value)
+
+    @property
+    def nested(self):
+        """Return value as nested list
+
+        For example the following entry is nested by default:
+        ##$VisuFGOrderDesc=( 2 )
+        (6, <FG_ECHO>, <>, 0, 1) (5, <FG_SLICE>, <>, 1, 2)
+
+        But this one not:
+        ##$VisuFGOrderDesc=( 2 )
+        (5, <FG_SLICE>, <>, 1, 2)
+
+        This proprerty allows treat the parameter as nested list in all cases.
+
+        """
+        value = self.list
+        if isinstance(value[0], list):
+            return value
+        else:
+            return [value]
+
+
+    @property
+    def array(self):
+        return np.atleast_1d(self.value)
+
+    @property
+    def shape(self):
+        value = self.value
+        if isinstance(value, np.ndarray):
+            return value.shape
+        else:
+            raise AttributeError
+
+
+
     @classmethod
     def pack_key(cls, value, usr_defined):
         assert isinstance(value, str)
@@ -120,7 +176,14 @@ class GenericParameter(Parameter):
 
     @property
     def value(self, **kwargs):
-        val_str_list = GenericParameter.split_parallel_lists(self.val_str)
+
+        val_str = re.sub('\n', '', self.val_str)
+
+        # unwrap wrapped list
+        if re.match('@[0-9]*\*',val_str) is not None:
+            val_str = self._unwrap_list(val_str)
+
+        val_str_list = GenericParameter.split_parallel_lists(val_str)
 
         if isinstance(val_str_list, str):
             value = GenericParameter.parse_value(val_str_list)
@@ -130,7 +193,10 @@ class GenericParameter(Parameter):
                 value.append(GenericParameter.parse_value(val_str))
 
         if isinstance(value, np.ndarray) and self.size:
-            return np.reshape(value, self.size, order='C')
+            if not 'str' in value.dtype.name:
+                return np.reshape(value, self.size, order='C')
+            else:
+                return value
         else:
             return value
 
@@ -156,6 +222,26 @@ class GenericParameter(Parameter):
         self.size = size
         self.val_str= val_str
 
+    def primed_dict(self, index):
+        nested_list = self.nested
+        primed_dict = OrderedDict()
+
+        for list in nested_list:
+            primed_dict[list[index]] = list
+
+        return primed_dict
+
+    def sub_list(self, index):
+        nested_list = self.nested
+        sub_list = []
+        for list in nested_list:
+            sub_list.append(list[index])
+
+        return sub_list
+
+
+
+
     @property
     def size(self):
         size_str = self.size_str[1:-2]
@@ -209,18 +295,26 @@ class GenericParameter(Parameter):
         val_str = re.sub('\n','', val_str)
 
         # sharp string
-        if val_str.startswith('<') and val_str.endswith('>') and '<' not in val_str[1:-1]:
-            return val_str
+        if val_str.startswith('<') and val_str.endswith('>'):
+
+            val_strs = re.findall('<[^<>]*>', val_str)
+
+            if len(val_strs) == 1:
+                return val_strs[0]
+            else:
+                return np.array(val_strs)
+
 
         # int/float
-        try:
-            value = ast.literal_eval(val_str)
+        if len(re.findall(GRAMMAR['SINGLE_NUMBER'],val_str))==1:
+            try:
+                value = ast.literal_eval(val_str)
 
-            # if value is int, or float, return, tuple will be parsed as list later on
-            if isinstance(value, float) or isinstance(value, int):
-                return value
-        except (ValueError, SyntaxError):
-            pass
+                # if value is int, or float, return, tuple will be parsed as list later on
+                if isinstance(value, float) or isinstance(value, int):
+                    return value
+            except (ValueError, SyntaxError):
+                pass
 
         # list
         if val_str.startswith('(') and val_str.endswith(''):
@@ -333,6 +427,22 @@ class GenericParameter(Parameter):
 
         return lst
 
+    def _unwrap_list(self, val_str):
+
+        while re.search('@[0-9]*\*\(\d*\.?\d*\)', val_str):
+            match = re.search('@[0-9]*\*\(\d*\.?\d*\)', val_str)
+            left = val_str[0:match.start()]
+            right = val_str[match.end():]
+            sub = val_str[match.start():match.end()]
+            size, value = re.split('\*', sub)
+            size = int(size[1:])
+            middle = ''
+            for i in range(size):
+                middle += '{} '.format(value[1:-1])
+            val_str = left + middle[0:-1] + right
+
+        return val_str
+
 
 class HeaderParameter(Parameter):
     def __init__(self, key_str, size_str, val_str, version):
@@ -363,6 +473,25 @@ class GeometryParameter(Parameter):
     def value(self):
         pass
 
+    @property
+    def affine(self):
+        """
+
+        :return: 4x4 3D Affine Transformation Matrix
+        """
+        # TODO support for multiple slice packages
+        match = re.match('\(\(\(.*\)', self.val_str)
+        affine_str = self.val_str[match.start() + 3: match.end() - 1]
+        orient, shift = affine_str.split(', ')
+
+        orient = GenericParameter.parse_value(orient)
+        shift = GenericParameter.parse_value(shift)
+        affine = np.zeros(shape=(4,4))
+        affine[0:3, 0:3] = np.reshape(orient, (3,3))
+        affine[0:3, 3] = shift
+
+        return affine
+
 
 class DataParameter(Parameter):
     def __init__(self, version, key, size_bracket, value):
@@ -456,7 +585,7 @@ class JCAMPDX(object):
 
             jcampdx_serial += '{}\n'.format(param_str)
 
-        return jcampdx_serial[0:-1] + "\nEND= "
+        return jcampdx_serial[0:-1] + "\n##END= "
 
     def __enter__(self):
         self.load()
@@ -470,7 +599,10 @@ class JCAMPDX(object):
         return self
 
     def __getitem__(self, key):
-        return self.get_value(key)
+        return self.params[key]
+
+    def __delitem__(self, key):
+        del self.params[key]
 
     def load(self):
         self.load_parameters()
@@ -511,6 +643,9 @@ class JCAMPDX(object):
     def version(self, value):
         self.version = value
 
+    def keys(self):
+        return self.params.keys()
+
     """
     PUBLIC INTERFACE
     """
@@ -559,6 +694,14 @@ class JCAMPDX(object):
     def get_float(self, key):
         return float(self.get_value(key))
 
+    def get_tuple(self, key):
+        value = self.get_value(key)
+
+        if isinstance(value, int) or isinstance(value, float):
+            return (value,)
+        else:
+            return tuple(value)
+
     def get_array(self, key, dtype=None, shape=(-1,), order='C'):
         parameter=self.get_parameter(key)
         value = parameter.value
